// Cree par David Croft pour le cours d'IA a l'ESI
// fevrier 2001

package edu.esi.lettresdespace.prof;

import edu.esi.lettresdespace.*;
import edu.esi.reseauxneuronaux.*;
import java.util.Vector;
import java.util.Enumeration;

/**
 * Ce programme donne la solution pour la probleme des lettres d'espace.
 */
public class SolutionApp {

	public static void main(String args[]) {

		// Cree une lecteur pour afficher les cadres video
		VideoLecteur lecteur = new VideoLecteurGraphique();
		
		// Charge le video 
		Video video = Video.charge("../video.dat");
		
		// Si nous voulons regarder tous les cadres dans la video -
		// mais pour une grande video ca prends trop de temps!
		/*
		for (int i=0; i < video.getSize(); i++) {
			Cadre next = video.getCadreAt(i);
			System.out.println("Cadre index: " + i);
			boolean response =  lecteur.confirmCadre(next);
		}
		*/
		
		// Prend le premier cadre pour connaitre la dimension pour la couche
		// de l'entree quand nous faisons le reseau.
		Cadre cadre = video.getCadreAt(0);
		
		// Cree le reseau avec 3 couche, remarque qu'il y a seulement un
		// noeuds pour la couche de la sortie
		int numCouches = 3;
		int[] numNoeudsPerCouche = { 	cadre.longueur(),	
										cadre.longueur()/4,
										1 };
		ReseauMulticouche reseau = 
			new ReseauMulticouche(	numCouches,
									numNoeudsPerCouche,
									false);

		// Imprime le reseau pour voir la structure
		System.out.println(reseau);
		System.out.println();
		
		// Charge les enchantillons avec quoi nous allons former le reseau
		System.out.println("Charge les enchantillons positifs...");
		Vector entrees = CadreChargeur.charge(	"../fichiers",
												"lettreA.dat");

		// Rappelle combine des enchantillons positifs
		int numPositifs = entrees.size();

		System.out.println("Charge les enchantillons negatifs...");
		Vector negatifs = CadreChargeur.charge(	"../fichiers",
												"toutesLettresSansA.dat");

		// ajoute tous les cadres a le meme Vector pour les entrees
		entrees.addAll(negatifs);
		
		// Cree les sorties (1.0 pour les positifs, 0.0 pour les negatifs)
		Vector sorties = new Vector(entrees.size());
		Double positif = new Double(1.0);
		Double negatif = new Double(0.0);

		for (int i=0; i < entrees.size(); i++) {
			VecteurESDouble vecteur = new VecteurESDouble(1);
			if (i < numPositifs) {
				vecteur.addElement(positif);
			} else {
				vecteur.addElement(negatif);
			}
			sorties.addElement(vecteur);
		}

		// Forme le reseau avec ces enchantillons - utilise beaucoup des
		// iterations, parce que la fonction va revenir quand l'erreur est
		// petit meme si nous n'avons pas fini tous les iterations.
		reseau.forme(entrees, sorties, 10000);
		
		// Cherche tous les cadres dans la video pour notre lettre
		// Remarque - j'utilise un seuil de 0.8 pour determiner si nous
		// allons afficher le cadre qui resemble possiblement notre lettre
		
		for (int i=0; i < video.getSize(); i++) {
			Cadre next = video.getCadreAt(i);

			// Donne le cadre au reseau, et regarde la production
			VecteurES output = reseau.reconnais(next);
			double value = output.getDoubleAt(0);
			if (value > 0.8) {
				System.out.println("Le cadre a index " + i +
					" donne une production de " + value);
				boolean response =  lecteur.confirmCadre(next);
				if (response == true) {
					System.out.println("Il y a une lettre A a index: " + i);
				}
			}
		}
		System.exit(0);
	}
}
